#ifndef _pvga1a_mmio_inc_INC_
#define _pvga1a_mmio_inc_INC_

#define VGA_MAXPAGE 16

#include "WD/pvga1a_io.h"


#define MM_PROT_RW (VM_READ|VM_WRITE|VM_SHARED)
#define MM_PROT_W  (VM_WRITE|VM_SHARED)
#define MM_PROT_R  (VM_READ|VM_SHARED)

/* Set the page to map and clear old mappings */
static void mmap_setpage_4bit(struct kgi_mmio_region *mm, kgi_uint offset)
{
	int page = (offset >> 16) & 0xF;

	DEBUG1("chipset_mmap_setpage: setting 4bit page to %d", page);
	outbGRC(page & 3, 4); /* Read Map select */
	outbGRC(page & 15, 0);/* Color select */

	mm->offset = (page << 16);
}

/* The PVGA1A and wd90c00 don't to ro/rw bank splitting, but the 
   current linmm_banked driver doesn't deal with this strategically,
   so we split into 32K banks and pretend they are ro/rw --
   inefficient but it gets the job done. */

static void mmap_setpage_linmm_abank (struct kgi_mmio_region *mm,kgi_uint offset)
{
        int page = (offset >> 12) & 0xF8;  /* no 4K granularity yet :-( */

	DEBUG1("chipset_mmap_setpage: setting PR0A (ascend) 4K page to %d", 
	       page);
	outPR0A((page - 8) & 0xFF); /* PVGA1A does _not_ adjust for the
                                       offset of the split window relative to
                                       the whole window (blech!) */
	mm->offset = (page << 12);
}

static void mmap_setpage_linmm_rbank (struct kgi_mmio_region *mm,kgi_uint offset)
{
        int page = (offset >> 12) & 0xF8;  /* no 4K granularity yet :-( */
	
	DEBUG1("chipset_mmap_setpage: setting PR0B (read-only) 4K page to %d", 
	       page);
	outPR0B(page & 0xFF);

	mm->offset = (page << 12);
}


static struct kgi_mmio_region mm_text16 = 
{
	NULL,			/* Next					*/
	"text16",		/* Name 				*/
	MMAP_FRAMEBUFFER,	/* Type (Linear)			*/
	MM_PROT_RW,		/* Protection map			*/
	0x10000,		/* Total size of area			*/
	0x10000,		/* Physical size (64k)			*/
	0xA0000,		/* Physical address of the windows	*/
	0,			/* Logical offset			*/
	NULL			/* Set offset function			*/
};
	
static struct kgi_mmio_region mm_1bit = 
{
	NULL,			/* Next					*/
	"VGA 1bit",		/* Name					*/
	MMAP_FRAMEBUFFER,	/* Type (Linear)			*/
	MM_PROT_RW,		/* Protection map			*/
	0x10000,		/* Total size of area			*/
	0x10000,		/* Physical size (64k)			*/
	0xA0000,		/* Physical address of the windows	*/
	0,			/* Logical offset			*/
	NULL			/* Set offset function			*/
};

static struct kgi_mmio_region mm_4bit = 
{
	NULL,			/* Next					*/
	"VGA 4bit",		/* Name 				*/
	MMAP_PRIVATE | 4,	/* Type (VGA 4bit planes)		*/
	MM_PROT_RW,		/* Protection map			*/
	16 * 0x10000,		/* Total size of area			*/
	0x10000,		/* Physical size (64k)			*/
	0xA0000,		/* Physical address of the windows	*/
	0,			/* Logical offset			*/
	mmap_setpage_4bit       /* Set offset function			*/
};
	
static struct kgi_mmio_region mm_8bit = 
{
	NULL,			/* Next					*/
	"8bit",			/* Name					*/
	MMAP_FRAMEBUFFER,	/* Type	(Linear)			*/
	MM_PROT_RW,		/* Protection map			*/
	0x10000,		/* Total size of area			*/
	0x10000,		/* Physical size (64k)			*/
	0xA0000,		/* Physical address of the windows	*/
	0,			/* Logical offset			*/
	NULL                    /* Set offset function			*/
};

static struct kgi_mmio_region mm_linmm_abank = 
{
	NULL,			/* Next					*/
	"Write Bank",       /* Name					*/
	MMAP_PRIVATE | 4,	/* Type MMAP_LINEAR_PAGED		*/
	MM_PROT_RW,		/* Protection map			*/
	0,		        /* Total size of area (mode dependent)  */
	0x08000,	  	/* Physical size (64k)			*/
	0xA8000,		/* Physical address of the windows	*/
	0,			/* Logical offset			*/
	mmap_setpage_linmm_abank /* Set offset function			*/
};

static struct kgi_mmio_region mm_linmm_rbank = 
{
	NULL,			/* Next					*/
	"Read Bank",            /* Name					*/
	MMAP_PRIVATE | 16,	/* Type MMAP_LINEAR_PAGED		*/
	MM_PROT_R,		/* Protection map			*/
	0,		        /* Total size of area (mode dependent)	*/
	0x08000,		/* Physical size (32k)			*/
	0xA0000,		/* Physical address of the windows	*/
	0,			/* Logical offset			*/
	mmap_setpage_linmm_rbank /* Set offset function			*/
};

static int graphics_mmio_setup(struct kgi_display *dpy, struct kgi_mode *mode)
{
	struct kgi_mmio_region *mm = NULL;

	switch (mode->request.graphtype) {
		case KGIGT_TEXT16:
		mm = &mm_text16;
		break;
		case KGIGT_1BIT:
		mm = &mm_1bit;
		break;
		case KGIGT_4BIT:
		mm = &mm_4bit;
		break;
		case KGIGT_8BIT:
		if (mode->request.virt.x <= 320 && 
		    mode->request.virt.y <= 200)
			mm = &mm_8bit;
		else
			mm = &mm_linmm_abank;
	                mm->next = &mm_linmm_rbank;
			mm->size = mm->next->size =
			  ((mode->request.virt.x * mode->request.virt.y) & 
			   ~(32 * 1024 - 1)) + 32 * 1024;
			/* Need better kernel mmap routines to do this:
			mm->size = mm->next->size =  
	                ((mode->request.virt.x * mode->request.virt.y) & 
			   ~(PAGE_SIZE - 1)) + PAGE_SIZE;
			   */

		break;
		default:
		return -E(ACCEL,BOGUS);
	}

	if (mm != NULL) {
	  if (mm->next != NULL) {
	    mm->next->next = mode->mmio;
	  }
	  else {
	    mm->next = mode->mmio;
	  }
	  mode->mmio = mm;
	}

	return EOK;
}

#endif /* _pvga1a_mmio_inc_ */
